#include "expr.h"
#include "dudloadroutines.h"
#include "expr_var.h"

#ifndef MAKEDEPEND
 #include <cmath>
#endif

class expr_sulkulause : public expr
{
	exprp content;
public:
	expr_sulkulause(const expr_sulkulause &s) : expr(s)
	{
		cout << __LINE__ << " in " __FILE__ ": This shouldn't happen!\n";
	}
	expr_sulkulause() { }
	virtual exprp load(string &input)
	{
		if(getchar(input) != '(')goto Invalid;
		
		content = expr::load(input);
		
		if(!content)
		Invalid:
			valid = false;
		else
		{
			blank(input);
			if(getchar(input) != ')')
				valid = false;
			else
			{
				type = content->type;
		}	}
		return this;
	}
	virtual exprp optimize()
	{
		return content;
	}
	virtual const string getname() const
	{
		return "(" + content->getname() + ")";
	}
	virtual exprp calculate(calculate_args) const
	{
		cout << "Warning: expr_sulkulause.calculate called\n";
		return content->calculate(calculate_call);
	}
};

// this define is for expr_2op.h
#define EXPR2OP_IMPLEMENT_FUNCTIONS

#include "expr_2op.h"

class expr_conditional : public expr
{
	exprp condition;
	exprp positive;
	exprp negative;
	
public:
	expr_conditional(const expr_conditional &s) : expr(s)
	{
		condition = s.condition;
		positive  = s.positive;
		negative  = s.negative;
	}
	expr_conditional(exprp l, string &input, int prio) : condition(l)
	{
		positive = expr::load(input, prio);
		if(!positive)
			valid = false;
		else
		{
			blank(input);
			if(getchar(input) != ':')
				valid = false;
			else
			{
				negative = expr::load(input, prio);
				if(!negative)valid = false;
			}
		}
	}	
	virtual exprp optimize()
	{
		if(valid && condition->isconst)return condition->truth() ? positive : negative;
		return this;
	}
	virtual const string getname() const
	{
		return "(" + condition->getname() + " ? " + positive->getname() + " : " + negative->getname() + ")";
	}
	virtual exprp calculate(calculate_args) const
	{
		exprp tmp = condition->calculate(calculate_call);		
		if(tmp->type == notsure)
		{
			return new expr_conditional(*this);
		}
		exprp which = tmp->truth() ? positive : negative;
		return which->calculate(calculate_call);
	}
};

class expr_booland : public expr
{
	exprp left;
	exprp right;
	
public:
	expr_booland(const expr_booland &s) : expr(s)
	{
		left = s.left;
		right= s.right;
	}
	expr_booland(exprp l, string &input, int prio) : left(l)
	{
		right = expr::load(input, prio);
		if(!right)valid = false;
	}
	virtual exprp optimize()
	{
		if(valid && left->isconst && right->isconst)
		{
			bool diu = left->truth();
			if(diu)diu = right->truth();
			return new expr_number(diu?1:0);
		}
		return this;
	}
	virtual const string getname() const
	{
		return "(" + left->getname() + "&&" + right->getname() + ")";
	}
	virtual exprp calculate(calculate_args) const
	{
		bool diu;
		exprp tmp;
		tmp = left->calculate(calculate_call);
		diu = tmp->truth();
		if(diu)
		{
			tmp = right->calculate(calculate_call);
			diu = tmp->truth();
		}
		return new expr_number(diu?1:0);
	}
};

class expr_boolor : public expr
{
	exprp left;
	exprp right;
	
public:
	expr_boolor(const expr_boolor &s) : expr(s)
	{
		left = s.left;
		right= s.right;
	}
	expr_boolor(exprp l, string &input, int prio) : left(l)
	{
		right = expr::load(input, prio);
		if(!right)valid = false;
	}
	
	virtual exprp optimize()
	{
		if(valid && left->isconst && right->isconst)
		{
			bool diu = left->truth();
			if(!diu)diu = right->truth();
			return new expr_number(diu?1:0);
		}
		return this;
	}
	virtual const string getname() const
	{
		return "(" + left->getname() + "||" + right->getname() + ")";
	}
	virtual exprp calculate(calculate_args) const
	{
		bool diu;
		exprp tmp;
		tmp = left->calculate(calculate_call);
		diu = left->truth();
		if(!diu)
		{
			tmp = right->calculate(calculate_call);
			diu = right->truth();
		}
		return new expr_number(diu?1:0);
	}
};


#include "expr_assign.h"
#include "expr_mathfun.h"
#include "expr_huuto.h"

#ifdef COMPET
extern expr_floattype X;
class expr_x : public expr
{
public:
	expr_x(string &)
	{
		type    = num;
		isconst = false;
	}
	virtual exprp calculate(calculate_args_notused) const
	{
		return new expr_number(X);
	}
	virtual const string getname() const
	{
		return "\33[1;35mX\33[m";
	}
	virtual const string getstr() const
	{
		return "X";
	}
};
#else
#include "expr_time.h"
#include "expr_ftime.h"
#ifdef DUD2
#include "expr_clone.h"
#endif
#endif

/*
 * prioriteetit:
 *
 *  0 = mik vain ky.
 *  1 = plus- ja miinuslaskujen sisll.
 *  2 = kerto- ja jakolaskun sisll. Ei saa breakata ykksell.
 *  3 = eksponenttitoimitus. Ei saa breikata pienemmll.
 *  4 = funktiossa. Ei saa breikata paitsi suluilla.
 *
 */
exprp expr::load(string &input, int prio)
{
	string s(getitem(input));
	
	exprp left;
	
	if(!s.size())
	{
		register char c = *input.c_str();	// Won't fail
		if(c=='-' || c=='+') // Negation, UGH. Plussa samalla vaivalla...
		{
			left = new expr_number(0);
			// Luomaan miinuslaskuoperaatio joka hakee toisen parametrin.
			goto moreop;
		}
		if(c != '(')		// Found invalid char.
		{
			// Expression needs to contain something.
			// It can't _begin_ with =, for example.
			return NULL;
		}
		left = (new expr_sulkulause())->load(input);
	}
	else if(s[0] == '"')
	{
		left = new expr_string(s.c_str()+1);
	}
	else if(!ischar(s[0]))
	{
		const char *c = s.c_str();
		expr_floattype d;
		if(*c=='0'&&c[1]=='x')
			d = expr_longfunc(c+2, NULL, 16); /* strtoul or similar */
		else
			d = expr_floatfunc(c, NULL);
		left = new expr_number(d);
	}
	else
	{
		#define expr_func(x) if(s==#x) left = (new expr_##x(input)); else
		
#ifdef COMPET
		expr_func(x)
#else
		expr_func(time)
		expr_func(ftime)
	#ifdef DUD2
		expr_func(clone)
	#endif		
#endif		
		expr_func(sin)
		expr_func(cos)
		expr_func(tan)
		expr_func(sqrt)
		expr_func(sinh)
		expr_func(cosh)
		expr_func(tanh)
		expr_func(exp)
		expr_func(log)
		expr_func(log10)
		expr_func(atan)
		expr_func(asin)
		expr_func(acos)
		expr_func(ceil)
		expr_func(floor)
		expr_func(atan2)
		left = new expr_variable(s); /* Note the "else" in expr_func */
		
		#undef expr_func
	}
	
moreop:
	if(!left || !left->valid)return NULL;
	
//	cout << "before = '" << left->getname() << "'\n";
	
	left = left->optimize();
	
//	cout << "after  = '" << left->getname() << "'\n";
//	cout << "input='" << input << "'\n";	

	blank(input);
	if(input.size())
	{
		if(prio < prio_additive)
		{
			if(input[0] == '+')
			{
				getchar(input);
				left = new expr_2op(left, input, prio_additive, expr_plus_fun);
				goto moreop;
			}
			if(input[0] == '-')
			{	
				getchar(input);
				left = new expr_2op(left, input, prio_additive, expr_minus_fun);
				goto moreop;
			}
		}
		if(prio < prio_equality)
		{
			if(input[0] == '<' || input[0] == '>' || input[0] == '=' || input[0] == '!')
			{
				string tmp = input;
				char c1 = getchar(input);
				char c2 = 0;
				if(input[0] == '<' || input[0] == '>' || input[0] == '=')c2 = getchar(input);
				
				if(c1=='=' && c2=='=')
				{
					left = new expr_2op(left, input, prio_equality, expr_eq_fun);
					goto moreop;
				}
				if((c1=='<' && c2=='>') || (c1=='<' && c2=='>') || (c1=='!' && c2=='='))
				{
					left = new expr_2op(left, input, prio_equality, expr_neq_fun);
					goto moreop;
				}
				input = tmp;
			}
		}
		if(prio < prio_boolean)
		{
			if(input[0] == '&')
			{
				getchar(input);
				if(input[0] == '&')getchar(input);
				left = new expr_booland(left, input, prio_boolean);
				goto moreop;
			}
			if(input[0] == '|')
			{
				getchar(input);
				if(input[0] == '|')getchar(input);
				left = new expr_boolor(left, input, prio_boolean);
				goto moreop;
			}
		}
		#ifdef DUD2
		if(prio < prio_assignment)
		{
			if(input[0] == '=' && input[1] != '=' && input[1] != '<' && input[1] != '>')
			{
				getchar(input);
				left = new expr_assign(left, input, prio_assignment);
				goto moreop;
			}
		}
		#endif
		if(prio < prio_relational)
		{
			if(input[0] == '<' || input[0] == '>' || input[0] == '=')
			{
				string tmp = input;
				char c1 = getchar(input);
				char c2 = 0;
				if(input[0] == '<' || input[0] == '>' || input[0] == '=')c2 = getchar(input);
				
				if(c1=='<' && !c2)
				{
					left = new expr_2op(left, input, prio_relational, expr_lt_fun);
					goto moreop;
				}
				if(c1=='>' && !c2)
				{
					left = new expr_2op(left, input, prio_relational, expr_gt_fun);
					goto moreop;
				}
				if((c1=='<' && c2=='=') || (c1=='=' && c2=='<'))
				{
					left = new expr_2op(left, input, prio_relational, expr_lteq_fun);
					goto moreop;
				}
				if((c1=='>' && c2=='=') || (c1=='=' && c2=='>'))
				{
					left = new expr_2op(left, input, prio_relational, expr_gteq_fun);
					goto moreop;
				}
				input = tmp;
			}
		}
		if(prio < prio_conditional)
		{
			if(input[0] == '?')
			{
				getchar(input);
				left = new expr_conditional(left, input, prio_conditional);
				goto moreop;
			}
		}
		if(prio < prio_divmul)
		{
			if(input[0] == '*')
			{
				getchar(input);
				left = new expr_2op(left, input, prio_divmul, expr_times_fun);
				goto moreop;
			}
			if(input[0] == '/')
			{
				getchar(input);
				left = new expr_2op(left, input, prio_divmul, expr_divided_fun);
				goto moreop;
			}
			if(input[0] == '%')
			{
				getchar(input);
				left = new expr_2op(left, input, prio_divmul, expr_fmod_fun);
				goto moreop;
			}
		}
		if(prio < prio_expon)
		{
			if(input[0] == '^')
			{	
				getchar(input);
				left = new expr_2op(left, input, prio_expon, expr_pow_fun);
				goto moreop;
			}
		}
		if(prio < prio_huuto)
		{
			if(input[0] == '!')
			{	
				getchar(input);
				left = new expr_huuto(left);
				goto moreop;
			}
		}
		// Any other character, like =, causes the
		// expression to terminate (without error)
	}
	
//	cout << "leaving = '" << left->getname() << "'\n";
	
	return left; /* Which is either new or null */
}

expr::~expr()
{
}
exprp expr::optimize()
{
	return this;
}

const string expr::getstr() const
{
	return "";
}
const string expr::getname() const
{
	return "?invalid_expression?";
}
const expr_floattype expr::getnum() const
{
	#ifdef __USE_ISOC9X
	return NAN;      // not a number
	#else
	return sqrt(-1);  // NaN we want.
	#endif
}
const bool expr::truth() const
{
	string s = getstr();
	const char *c = s.c_str();
	
	while(*c=='0')c++;
	return *c ? true : false;
}
exprp expr::calculate(calculate_args_notused) const
{
	cout << "Fatal: Undefined object (content: '" << getname() << "') tried to calculate itself.\n";
	return NULL;
}
#ifdef DUD2
obj::vartype &expr::getvar(calculate_args) const
{
	cout << "Warning: Invalid variable referenced\n";
	return thisobj->vars["#invalid"];
}
obj *expr::getobj(calculate_args) const
{
	return NULL;
}
#endif
